@@ -27,8 +27,9 @@
  * implementations. Subclasses should usually pass the filters they
  * want and override the D-Bus methods they implement.
  *
- * See #TpSimpleObserver for a class implementing a simple observer using
- * #TpBaseClient.
+ * For many applications, the provided #TpSimpleObserver, #TpSimpleApprover
+ * and #TpSimpleHandler subclasses can be used instead of deriving from this
+ * class.
  */
 
 /**
@@ -166,6 +167,7 @@
 #include <dbus/dbus-glib-lowlevel.h>
 
 #include <telepathy-glib/add-dispatch-operation-context-internal.h>
+#include <telepathy-glib/automatic-proxy-factory.h>
 #include <telepathy-glib/channel-dispatch-operation-internal.h>
 #include <telepathy-glib/channel-request.h>
 #include <telepathy-glib/channel.h>
@@ -202,6 +204,7 @@ enum {
     PROP_ACCOUNT_MANAGER,
     PROP_NAME,
     PROP_UNIQUIFY_NAME,
+    PROP_CHANNEL_FACTORY,
     N_PROPS
 };
 
@@ -222,6 +225,7 @@ typedef enum {
     CLIENT_HANDLER_WANTS_REQUESTS = 1 << 3,
     CLIENT_HANDLER_BYPASSES_APPROVAL = 1 << 4,
     CLIENT_OBSERVER_RECOVER = 1 << 5,
+    CLIENT_OBSERVER_DELAYS_APPROVERS = 1 << 6,
 } ClientFlags;
 
 struct _TpBaseClientPrivate
@@ -253,6 +257,7 @@ struct _TpBaseClientPrivate
 
   TpAccountManager *account_mgr;
   TpAccount *only_for_account;
+  TpClientChannelFactory *channel_factory;
 
   /* array of GQuark or NULL */
   GArray *account_features;
@@ -406,7 +411,56 @@ tp_base_client_set_observer_recover (TpBaseClient *self,
   g_return_if_fail (!self->priv->registered);
   g_return_if_fail (cls->observe_channels != NULL);
 
-  self->priv->flags |= (CLIENT_IS_OBSERVER | CLIENT_OBSERVER_RECOVER);
+  if (recover)
+    {
+      self->priv->flags |= (CLIENT_IS_OBSERVER | CLIENT_OBSERVER_RECOVER);
+    }
+  else
+    {
+      self->priv->flags |= CLIENT_IS_OBSERVER;
+      self->priv->flags &= ~CLIENT_OBSERVER_RECOVER;
+    }
+}
+
+/**
+ * tp_base_client_set_observer_delay_approvers:
+ * @self: a #TpBaseClient
+ * @delay: the value of the Observer.DelayApprovers property
+ *
+ * Set whether the channel dispatcher should wait for
+ * tp_observe_channels_context_accept() or tp_observe_channels_context_fail()
+ * to be called before calling
+ * #TpBaseClientClass.add_dispatch_operation on appropriate Approvers.
+ *
+ * This is implemented by setting the value of the DelayApprovers
+ * D-Bus property.
+ *
+ * This method may only be called before tp_base_client_register() is
+ * called, and may only be called on objects whose class implements
+ * #TpBaseClientClass.observe_channels.
+ *
+ * Since: 0.13.16
+ */
+void
+tp_base_client_set_observer_delay_approvers (TpBaseClient *self,
+    gboolean delay)
+{
+  TpBaseClientClass *cls = TP_BASE_CLIENT_GET_CLASS (self);
+
+  g_return_if_fail (TP_IS_BASE_CLIENT (self));
+  g_return_if_fail (!self->priv->registered);
+  g_return_if_fail (cls->observe_channels != NULL);
+
+  if (delay)
+    {
+      self->priv->flags |= (CLIENT_IS_OBSERVER |
+          CLIENT_OBSERVER_DELAYS_APPROVERS);
+    }
+  else
+    {
+      self->priv->flags |= CLIENT_IS_OBSERVER;
+      self->priv->flags &= ~CLIENT_OBSERVER_DELAYS_APPROVERS;
+    }
 }
 
 /**
@@ -473,12 +527,12 @@ tp_base_client_take_approver_filter (TpBaseClient *self,
  * tp_base_client_be_a_handler:
  * @self: a #TpBaseClient
  *
- * Register @self as a ChannelHandler with an empty list of filter.
+ * Register @self as a Client.Handler with an empty list of filters.
  * This is useful if you want to create a client that only handle channels
  * for which it's the PreferredHandler.
  *
  * This method may only be called before tp_base_client_register() is
- * called, and may only be called on objects whose class has called
+ * called, and may only be called on objects whose class implements
  * #TpBaseClientClass.handle_channels.
  *
  * Since: 0.11.6
@@ -599,10 +653,10 @@ tp_base_client_set_handler_bypass_approval (TpBaseClient *self,
  * @self: a #TpBaseClient
  *
  * Indicate that @self is a Handler willing to be notified about requests for
- * channels that it is likely to be asked to handle.
- * That means the TpBaseClient::request-added and TpBaseClient::request-removed:
- * signals will be fired and tp_base_client_get_pending_requests() will
- * return the list of pending requests.
+ * channels that it is likely to be asked to handle. This means that the
+ * #TpBaseClient::request-added and #TpBaseClient::request-removed signals will
+ * be fired and tp_base_client_get_pending_requests() will return the list of
+ * pending requests.
  *
  * This method may only be called before tp_base_client_register() is
  * called, and may only be called on objects whose class implements
@@ -907,6 +961,7 @@ tp_base_client_dispose (GObject *object)
   tp_clear_object (&self->priv->dbus);
   tp_clear_object (&self->priv->account_mgr);
   tp_clear_object (&self->priv->only_for_account);
+  tp_clear_object (&self->priv->channel_factory);
 
   g_list_foreach (self->priv->pending_requests, (GFunc) g_object_unref, NULL);
   g_list_free (self->priv->pending_requests);
@@ -914,7 +969,8 @@ tp_base_client_dispose (GObject *object)
 
   if (self->priv->my_chans != NULL &&
       g_hash_table_size (self->priv->my_chans) > 0)
-    WARNING ("TpBaseClient is still handling channels at dispose");
+    WARNING ("TpBaseClient is still handling %d channels at dispose",
+        g_hash_table_size (self->priv->my_chans));
 
   tp_clear_pointer (&self->priv->my_chans, g_hash_table_unref);
 
@@ -973,6 +1029,10 @@ tp_base_client_get_property (GObject *object,
         g_value_set_boolean (value, self->priv->uniquify_name);
         break;
 
+      case PROP_CHANNEL_FACTORY:
+        g_value_set_object (value, self->priv->channel_factory);
+        break;
+
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
         break;
@@ -1008,6 +1068,11 @@ tp_base_client_set_property (GObject *object,
         self->priv->uniquify_name = g_value_get_boolean (value);
         break;
 
+      case PROP_CHANNEL_FACTORY:
+        g_assert (self->priv->channel_factory == NULL);    /* construct-only */
+        self->priv->channel_factory = g_value_dup_object (value);
+        break;
+
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
         break;
@@ -1074,6 +1139,16 @@ tp_base_client_constructed (GObject *object)
   g_strdelimit (self->priv->object_path, ".", '/');
 
   self->priv->bus_name = g_string_free (string, FALSE);
+
+  if (self->priv->channel_factory == NULL)
+    {
+      self->priv->channel_factory = TP_CLIENT_CHANNEL_FACTORY (
+          tp_automatic_proxy_factory_dup ());
+    }
+  else
+    {
+      g_assert (TP_IS_CLIENT_CHANNEL_FACTORY (self->priv->channel_factory));
+    }
 }
 
 typedef enum {
@@ -1085,6 +1160,7 @@ typedef enum {
     DP_HANDLED_CHANNELS,
     DP_OBSERVER_CHANNEL_FILTER,
     DP_OBSERVER_RECOVER,
+    DP_OBSERVER_DELAY_APPROVERS,
 } ClientDBusProp;
 
 static void
@@ -1163,6 +1239,11 @@ tp_base_client_get_dbus_properties (GObject *object,
           (self->priv->flags & CLIENT_OBSERVER_RECOVER) != 0);
       break;
 
+    case DP_OBSERVER_DELAY_APPROVERS:
+      g_value_set_boolean (value,
+          (self->priv->flags & CLIENT_OBSERVER_DELAYS_APPROVERS) != 0);
+      break;
+
     default:
       g_assert_not_reached ();
     }
@@ -1197,6 +1278,8 @@ tp_base_client_class_init (TpBaseClientClass *cls)
           GINT_TO_POINTER (DP_OBSERVER_CHANNEL_FILTER) },
         { "Recover",
           GINT_TO_POINTER (DP_OBSERVER_RECOVER) },
+        { "DelayApprovers",
+          GINT_TO_POINTER (DP_OBSERVER_DELAY_APPROVERS) },
         { NULL }
   };
   static TpDBusPropertiesMixinIfaceImpl prop_ifaces[] = {
@@ -1309,6 +1392,25 @@ tp_base_client_class_init (TpBaseClientClass *cls)
   g_object_class_install_property (object_class, PROP_UNIQUIFY_NAME,
       param_spec);
 
+  /**
+   * TpBaseClient:channel-factory:
+   *
+   * The object implementing the #TpClientChannelFactoryInterface interface
+   * that will be used to create channel proxies. While
+   * tp_base_client_register() has not yet been called, this property can be
+   * changed using tp_base_client_set_channel_factory().
+   *
+   * If no channel factory is specified then #TpAutomaticProxyFactory is used.
+   *
+   * Since: 0.13.2
+   */
+  param_spec = g_param_spec_object ("channel-factory", "Channel factory",
+      "Object implementing TpClientChannelFactoryInterface",
+      G_TYPE_OBJECT,
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_CHANNEL_FACTORY,
+      param_spec);
+
  /**
    * TpBaseClient::request-added:
    * @self: a #TpBaseClient
@@ -1429,6 +1531,28 @@ array_data_or_null (GArray *array)
     return array->data;
 }
 
+static GArray *
+get_features_for_channel (TpBaseClient *self,
+    TpChannel *channel)
+{
+  GArray *features;
+
+  features = tp_client_channel_factory_dup_channel_features (
+      self->priv->channel_factory, channel);
+
+  g_assert (features != NULL);
+
+  /* Add TpBaseClient's own features, if any */
+  if (self->priv->channel_features == NULL)
+    return features;
+
+  _tp_quark_array_merge (features,
+      (GQuark *) self->priv->channel_features->data,
+      self->priv->channel_features->len);
+
+  return features;
+}
+
 static void
 _tp_base_client_observe_channels (TpSvcClientObserver *iface,
     const gchar *account_path,
@@ -1448,6 +1572,8 @@ _tp_base_client_observe_channels (TpSvcClientObserver *iface,
   GPtrArray *channels = NULL, *requests = NULL;
   TpChannelDispatchOperation *dispatch_operation = NULL;
   guint i;
+  TpChannel *channel = NULL;
+  GArray *channel_features;
 
   if (!(self->priv->flags & CLIENT_IS_OBSERVER))
     {
@@ -1495,13 +1621,13 @@ _tp_base_client_observe_channels (TpSvcClientObserver *iface,
     {
       const gchar *chan_path;
       GHashTable *chan_props;
-      TpChannel *channel;
 
       tp_value_array_unpack (g_ptr_array_index (channels_arr, i), 2,
           &chan_path, &chan_props);
 
-      channel = tp_channel_new_from_properties (connection,
-          chan_path, chan_props, &error);
+      channel = tp_client_channel_factory_create_channel (
+          self->priv->channel_factory, connection, chan_path, chan_props,
+          &error);
       if (channel == NULL)
         {
           DEBUG ("Failed to create TpChannel: %s", error->message);
@@ -1548,13 +1674,16 @@ _tp_base_client_observe_channels (TpSvcClientObserver *iface,
   ctx = _tp_observe_channels_context_new (account, connection, channels,
       dispatch_operation, requests, observer_info, context);
 
+  channel_features = get_features_for_channel (self, channel);
+
   _tp_observe_channels_context_prepare_async (ctx,
       array_data_or_null (self->priv->account_features),
       array_data_or_null (self->priv->connection_features),
-      array_data_or_null (self->priv->channel_features),
+      array_data_or_null (channel_features),
       context_prepare_cb, self);
 
   g_object_unref (ctx);
+  g_array_free (channel_features, TRUE);
 
 out:
   if (channels != NULL)
@@ -1645,6 +1774,8 @@ _tp_base_client_add_dispatch_operation (TpSvcClientApprover *iface,
   TpChannelDispatchOperation *dispatch_operation = NULL;
   guint i;
   const gchar *path;
+  TpChannel *channel = NULL;
+  GArray *channel_features;
 
   if (!(self->priv->flags & CLIENT_IS_APPROVER))
     {
@@ -1709,13 +1840,13 @@ _tp_base_client_add_dispatch_operation (TpSvcClientApprover *iface,
     {
       const gchar *chan_path;
       GHashTable *chan_props;
-      TpChannel *channel;
 
       tp_value_array_unpack (g_ptr_array_index (channels_arr, i), 2,
           &chan_path, &chan_props);
 
-      channel = tp_channel_new_from_properties (connection,
-          chan_path, chan_props, &error);
+      channel = tp_client_channel_factory_create_channel (
+          self->priv->channel_factory, connection, chan_path, chan_props,
+          &error);
       if (channel == NULL)
         {
           DEBUG ("Failed to create TpChannel: %s", error->message);
@@ -1737,13 +1868,16 @@ _tp_base_client_add_dispatch_operation (TpSvcClientApprover *iface,
   ctx = _tp_add_dispatch_operation_context_new (account, connection, channels,
       dispatch_operation, context);
 
+  channel_features = get_features_for_channel (self, channel);
+
   _tp_add_dispatch_operation_context_prepare_async (ctx,
       array_data_or_null (self->priv->account_features),
       array_data_or_null (self->priv->connection_features),
-      array_data_or_null (self->priv->channel_features),
+      array_data_or_null (channel_features),
       add_dispatch_context_prepare_cb, self);
 
   g_object_unref (ctx);
+  g_array_free (channel_features, TRUE);
 
 out:
   if (channels != NULL)
@@ -1777,10 +1911,14 @@ chan_invalidated_cb (TpChannel *channel,
     gchar *message,
     TpBaseClient *self)
 {
-  DEBUG ("Channel %s has been invalidated", tp_proxy_get_object_path (channel));
+  DEBUG ("Channel (%p) %s has been invalidated (%s)",
+      channel, tp_proxy_get_object_path (channel), message);
 
-  g_hash_table_remove (self->priv->my_chans, tp_proxy_get_object_path (
-        channel));
+  if (!(domain == TP_DBUS_ERRORS && code == TP_DBUS_ERROR_PROXY_UNREFERENCED))
+    {
+      g_hash_table_remove (self->priv->my_chans, tp_proxy_get_object_path (
+          channel));
+    }
 }
 
 static void
@@ -1795,7 +1933,9 @@ ctx_done_cb (TpHandleChannelsContext *context,
 
       if (tp_proxy_get_invalidated (channel) == NULL)
         {
-          g_hash_table_insert (self->priv->my_chans,
+          DEBUG ("Inserting Channel (%p) %s",
+            channel, tp_proxy_get_object_path (channel));
+          g_hash_table_replace (self->priv->my_chans,
               (gchar *) tp_proxy_get_object_path (channel),
               g_object_ref (channel));
 
@@ -1886,6 +2026,8 @@ _tp_base_client_handle_channels (TpSvcClientHandler *iface,
   TpConnection *connection = NULL;
   GPtrArray *channels = NULL, *requests = NULL;
   guint i;
+  TpChannel *channel = NULL;
+  GArray *channel_features;
 
   if (!(self->priv->flags & CLIENT_IS_HANDLER))
     {
@@ -1930,13 +2072,13 @@ _tp_base_client_handle_channels (TpSvcClientHandler *iface,
     {
       const gchar *chan_path;
       GHashTable *chan_props;
-      TpChannel *channel;
 
       tp_value_array_unpack (g_ptr_array_index (channels_arr, i), 2,
           &chan_path, &chan_props);
 
-      channel = tp_channel_new_from_properties (connection,
-          chan_path, chan_props, &error);
+      channel = tp_client_channel_factory_create_channel (
+          self->priv->channel_factory, connection, chan_path, chan_props,
+          &error);
       if (channel == NULL)
         {
           DEBUG ("Failed to create TpChannel: %s", error->message);
@@ -1975,13 +2117,16 @@ _tp_base_client_handle_channels (TpSvcClientHandler *iface,
   ctx = _tp_handle_channels_context_new (account, connection, channels,
       requests, user_action_time, handler_info, context);
 
+  channel_features = get_features_for_channel (self, channel);
+
   _tp_handle_channels_context_prepare_async (ctx,
       array_data_or_null (self->priv->account_features),
       array_data_or_null (self->priv->connection_features),
-      array_data_or_null (self->priv->channel_features),
+      array_data_or_null (channel_features),
       handle_channels_context_prepare_cb, self);
 
   g_object_unref (ctx);
+  g_array_free (channel_features, TRUE);
 
 out:
   if (channels != NULL)
@@ -2487,8 +2632,10 @@ tp_base_client_add_channel_features_varargs (TpBaseClient *self,
  *
  * Request that the given features are prepared on each #TpAccount (in
  * addition to %TP_ACCOUNT_FEATURE_CORE) before calling
- * #TpBaseClient.observe_channels, #TpBaseClient.add_dispatch_operation or
- * #TpBaseClient.handle_channels, or emitting #TpBaseClient::request-added.
+ * #TpBaseClientClass.observe_channels,
+ * #TpBaseClientClass.add_dispatch_operation or
+ * #TpBaseClientClass.handle_channels, or emitting
+ * #TpBaseClient::request-added.
  *
  * Since: 0.11.14
  */
@@ -2514,8 +2661,9 @@ tp_base_client_add_account_features (TpBaseClient *self,
  *
  * Request that the given features are prepared on each #TpChannel (in
  * addition to %TP_CHANNEL_FEATURE_CORE) before calling
- * #TpBaseClient.observe_channels, #TpBaseClient.add_dispatch_operation or
- * #TpBaseClient.handle_channels.
+ * #TpBaseClientClass.observe_channels,
+ * #TpBaseClientClass.add_dispatch_operation or
+ * #TpBaseClientClass.handle_channels.
  *
  * Since: 0.11.14
  */
@@ -2541,8 +2689,9 @@ tp_base_client_add_channel_features (TpBaseClient *self,
  *
  * Request that the given features are prepared on each #TpConnection (in
  * addition to %TP_CONNECTION_FEATURE_CORE) before calling
- * #TpBaseClient.observe_channels, #TpBaseClient.add_dispatch_operation or
- * #TpBaseClient.handle_channels.
+ * #TpBaseClientClass.observe_channels,
+ * #TpBaseClientClass.add_dispatch_operation or
+ * #TpBaseClientClass.handle_channels.
  *
  * Since: 0.11.14
  */
@@ -2560,3 +2709,45 @@ tp_base_client_add_connection_features (TpBaseClient *self,
 
   _tp_quark_array_merge (self->priv->connection_features, features, n);
 }
+
+/**
+ * tp_base_client_set_channel_factory:
+ * @self: a #TpBaseClient
+ * @factory: an object implementing the #TpClientChannelFactoryInterface
+ * interface
+ *
+ * Change the value of the #TpBaseClient:channel-factory property.
+ * It can't be changed once @self has been registered.
+ *
+ * Since: 0.13.2
+ */
+void
+tp_base_client_set_channel_factory (TpBaseClient *self,
+    TpClientChannelFactory *factory)
+{
+  g_return_if_fail (TP_IS_BASE_CLIENT (self));
+  g_return_if_fail (!self->priv->registered);
+  g_return_if_fail (TP_IS_CLIENT_CHANNEL_FACTORY (factory));
+
+  tp_clear_object (&self->priv->channel_factory);
+
+  self->priv->channel_factory = g_object_ref (factory);
+  g_object_notify (G_OBJECT (self), "channel-factory");
+}
+
+/**
+ * tp_base_client_get_channel_factory: (skip)
+ * @self: a #TpBaseClient
+ *
+ * Return the #TpBaseClient:channel-factory property.
+ *
+ * Returns: the value of #TpBaseClient:channel-factory
+ * Since: 0.13.2
+ */
+TpClientChannelFactory *
+tp_base_client_get_channel_factory (TpBaseClient *self)
+{
+  g_return_val_if_fail (TP_IS_BASE_CLIENT (self), NULL);
+
+  return self->priv->channel_factory;
+}
