@@ -29,11 +29,12 @@
 G_BEGIN_DECLS
 
 GQuark tp_errors_quark (void);
+GQuark tp_error_quark (void);
 
 #define TP_ERROR_PREFIX "org.freedesktop.Telepathy.Error"
 
-#define TP_ERRORS (tp_errors_quark ())
-#define TP_ERROR TP_ERRORS
+#define TP_ERRORS TP_ERROR
+#define TP_ERROR (tp_error_quark ())
 
 void tp_g_set_error_invalid_handle_type (guint type, GError **error);
 void tp_g_set_error_unsupported_handle_type (guint type, GError **error);
@@ -78,6 +79,14 @@ typedef enum {
     TP_ERROR_SERVICE_BUSY, /*< nick=ServiceBusy >*/
     TP_ERROR_RESOURCE_UNAVAILABLE, /*< nick=ResourceUnavailable >*/
     TP_ERROR_WOULD_BREAK_ANONYMITY, /*< nick=WouldBreakAnonymity >*/
+    TP_ERROR_CERT_REVOKED, /*< nick=Cert.Revoked >*/
+    TP_ERROR_CERT_INSECURE, /*< nick=Cert.Insecure >*/
+    TP_ERROR_CERT_LIMIT_EXCEEDED, /*< nick=Cert.LimitExceeded >*/
+    TP_ERROR_NOT_YET, /*< nick=NotYet >*/
+    TP_ERROR_REJECTED, /*< nick=Rejected >*/
+    TP_ERROR_PICKED_UP_ELSEWHERE, /*< nick=PickedUpElsewhere >*/
+    TP_ERROR_CONFUSED, /*< nick=Confused >*/
+    TP_ERROR_SERVICE_CONFUSED, /*< nick=ServiceConfused >*/
 } TpError;
 
 const gchar *tp_error_get_dbus_name (TpError error);
