@@ -2,6 +2,8 @@
  * util.c - Source for telepathy-glib utility functions
  * Copyright © 2006-2010 Collabora Ltd. <http://www.collabora.co.uk/>
  * Copyright © 2006-2008 Nokia Corporation
+ * Copyright © 1999 Tom Tromey
+ * Copyright © 2000 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -27,11 +29,19 @@
  * GLib, but aren't.
  */
 
+#include <config.h>
+
 #include <gobject/gvaluecollector.h>
 
+#ifdef HAVE_GIO_UNIX
+#include <gio/gunixsocketaddress.h>
+#include <gio/gunixconnection.h>
+#endif /* HAVE_GIO_UNIX */
+
 #include <telepathy-glib/util-internal.h>
 #include <telepathy-glib/util.h>
 
+#include <stdio.h>
 #include <string.h>
 
 #define DEBUG_FLAG TP_DEBUG_MISC
@@ -587,7 +597,11 @@ tp_mixin_offset_cast (gpointer instance, guint offset)
  *
  * If the type of @instance, or any of its ancestor types, has had an offset
  * attached using qdata with the given @quark, return that offset. If not,
- * this indicates a programming error and results are undefined.
+ * return 0.
+ *
+ * In older telepathy-glib versions, calling this function on an instance that
+ * did not have the mixin was considered to be a programming error. Since
+ * version 0.13.9, 0 is returned, without error.
  *
  * This is used to implement the telepathy-glib mixin classes.
  *
@@ -609,7 +623,7 @@ tp_mixin_instance_get_offset (gpointer instance,
         return GPOINTER_TO_UINT (qdata);
     }
 
-  g_return_val_if_reached (0);
+  return 0;
 }
 
 
@@ -619,8 +633,12 @@ tp_mixin_instance_get_offset (gpointer instance,
  * @quark: A quark that was used to store the offset with g_type_set_qdata()
  *
  * If the type of @klass, or any of its ancestor types, has had an offset
- * attached using qdata with the given @quark, return that offset. If not,
- * this indicates a programming error and results are undefined.
+ * attached using qdata with the given @quark, return that offset; if not,
+ * return 0.
+ *
+ * In older telepathy-glib versions, calling this function on an instance that
+ * did not have the mixin was considered to be a programming error. Since
+ * version 0.13.9, 0 is returned, without error.
  *
  * This is used to implement the telepathy-glib mixin classes.
  *
@@ -642,7 +660,7 @@ tp_mixin_class_get_offset (gpointer klass,
         return GPOINTER_TO_UINT (qdata);
     }
 
-  g_return_val_if_reached (0);
+  return 0;
 }
 
 
@@ -977,6 +995,8 @@ tp_g_signal_connect_object (gpointer instance,
   g_return_val_if_fail (detailed_signal != NULL, 0);
   g_return_val_if_fail (c_handler != NULL, 0);
   g_return_val_if_fail (G_IS_OBJECT (gobject), 0);
+  g_return_val_if_fail (
+      (connect_flags & ~(G_CONNECT_AFTER|G_CONNECT_SWAPPED)) == 0, 0);
 
   if (connect_flags & G_CONNECT_SWAPPED)
     ctx->closure = g_cclosure_new_object_swap (c_handler, gobject);
@@ -1490,3 +1510,143 @@ next_i:
       continue;
     }
 }
+
+#ifdef HAVE_GIO_UNIX
+GSocketAddress *
+_tp_create_temp_unix_socket (GSocketService *service,
+    GError **error)
+{
+  guint i;
+  GSocketAddress *address;
+
+  /* why doesn't GIO provide a method to create a socket we don't
+   * care about? Iterate until we find a valid temporary name.
+   *
+   * Try a maximum of 10 times to get a socket */
+  for (i = 0; i < 10; i++)
+    {
+      address = g_unix_socket_address_new (tmpnam (NULL));
+
+      g_clear_error (error);
+
+      if (g_socket_listener_add_address (
+            G_SOCKET_LISTENER (service),
+            address, G_SOCKET_TYPE_STREAM,
+            G_SOCKET_PROTOCOL_DEFAULT,
+            NULL, NULL, error))
+        return address;
+      else
+        g_object_unref (address);
+    }
+
+  return NULL;
+}
+#endif /* HAVE_GIO_UNIX */
+
+GList *
+_tp_create_channel_request_list (TpDBusDaemon *dbus,
+    GHashTable *request_props)
+{
+  GHashTableIter iter;
+  GList *result = NULL;
+  gpointer key, value;
+
+  g_hash_table_iter_init (&iter, request_props);
+  while (g_hash_table_iter_next (&iter, &key, &value))
+    {
+      TpChannelRequest *req;
+      const gchar *path = key;
+      GHashTable *props = value;
+      GError *error = NULL;
+
+      req = tp_channel_request_new (dbus, path, props, &error);
+      if (req == NULL)
+        {
+          DEBUG ("Failed to create TpChannelRequest: %s", error->message);
+          g_error_free (error);
+          continue;
+        }
+
+      result = g_list_prepend (result, req);
+    }
+
+  return result;
+}
+
+/**
+ * tp_utf8_make_valid:
+ * @name: string to coerce into UTF8
+ *
+ * Validate that the provided string is valid UTF8. If not,
+ * replace all invalid bytes with unicode replacement
+ * character (U+FFFD).
+ *
+ * This method is a verbatim copy of glib's internal
+ * _g_utf8_make_valid() function, and will be deprecated as
+ * soon as the glib one becomes public.
+ *
+ * Returns: a new valid UTF8 string
+ *
+ * Since: 0.13.15
+ */
+gchar *
+tp_utf8_make_valid (const gchar *name)
+{
+  GString *string;
+  const gchar *remainder, *invalid;
+  gint remaining_bytes, valid_bytes;
+
+  g_return_val_if_fail (name != NULL, NULL);
+
+  string = NULL;
+  remainder = name;
+  remaining_bytes = strlen (name);
+
+  while (remaining_bytes != 0)
+    {
+      if (g_utf8_validate (remainder, remaining_bytes, &invalid))
+        break;
+      valid_bytes = invalid - remainder;
+
+      if (string == NULL)
+        string = g_string_sized_new (remaining_bytes);
+
+      g_string_append_len (string, remainder, valid_bytes);
+      /* append U+FFFD REPLACEMENT CHARACTER */
+      g_string_append (string, "\357\277\275");
+
+      remaining_bytes -= valid_bytes + 1;
+      remainder = invalid + 1;
+    }
+
+  if (string == NULL)
+    return g_strdup (name);
+
+  g_string_append (string, remainder);
+
+  g_assert (g_utf8_validate (string->str, -1, NULL));
+
+  return g_string_free (string, FALSE);
+}
+
+gboolean
+_tp_bind_connection_status_to_boolean (GBinding *binding,
+    const GValue *src_value,
+    GValue *dest_value,
+    gpointer user_data)
+{
+  gboolean invert = GPOINTER_TO_UINT (user_data);
+  gboolean value;
+
+  g_return_val_if_fail (G_VALUE_HOLDS_UINT (src_value), FALSE);
+  g_return_val_if_fail (G_VALUE_HOLDS_BOOLEAN (dest_value), FALSE);
+
+  value = (g_value_get_uint (src_value) == TP_CONNECTION_STATUS_CONNECTED);
+
+  if (invert)
+    value = !value;
+
+  g_value_set_boolean (dest_value, value);
+
+  return TRUE;
+}
